      ////////////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM         //
    //  (Version 2)                                ///
   // Written by Andrew Broad                     ////
  //                                             /////
 // Frame.java (2001:09:20 16:24)               //////
//////////////////////////////////////////////////////
// Methods of Frame:                            //////
// - static initialise                          //////
// - Frame() constructor                        //////
// - Frame(FrameClass) constructor              //////
// - Frame(String) constructor                  //////
// - getFrameClass                              //////
// - getNumber                                  //////
// - setNumber                                  //////
// - getSlotFiller(Slot)                        //////
// - getSlotFiller(String)                      //////
// - gsf                                        //////
// - setSlotFiller(Slot,Object)                 //////
// - setInverseSlot                             //////
// - setSlotFillerIndirectly                    //////
// - setSlotFiller(String,Object)               //////
// - ssf                                        //////
// - addElementToSlotFiller(Slot,Object)        //////
// - addElementToSlotFiller(String,Object)      //////
// - aesf                                       //////
// - deleteElementFromSlotFiller(Slot,Object)   //////
// - deleteElementFromSlotFiller(String,Object) //////
// - desf                                       //////
// - isSlotSetDirectly                          //////
// - getSlots                                   //////
// - markCaseProblemPart                        //////
// - isCaseProblemPart                          //////
// - isInputModelFrame                          //////
// - toString                                   //////
// - print                                      //////
// - static generateFIL                         /////
// - generateFIL                                ////
// - FILise                                     ///
// - static remove                              //
/////////////////////////////////////////////////

import java.io.*;
import java.util.*;

import jex.*;

/*****************************************************************************
** The class `Frame' represents a frame (an instance of a
** <A HREF="FrameClass.html">FrameClass</A>).
*****************************************************************************/

public class Frame implements FrameInterface, Serializable, FILisable {
	public static boolean _markFramesAsInputModel;
	public static int _numberOfFrames;

	private OrderedHashtable/*Slot->Object*/ _frameRep
	= new OrderedHashtable();
	private Hashtable/*Slot->boolean*/ _slotsSetDirectly= new Hashtable();
	private FrameClass _frameClass;
	protected int _number;
	private boolean _caseProblemPart= false;
	private boolean _fromInputModel;

	/********************************************************************/
	// static Frame.initialise : void -> void
	/*********************************************************************
	** Initialises the Jex subclasses of Frame, after loading the
	** EXPRESS metamodel from EXPRESS/frame_classes.exp.
	*********************************************************************/

	public static void initialise () {
		//////////////////////////////////////////////////////////////
		// Parse the EXPRESS metamodel in EXPRESS/frame_classes.exp //
		//////////////////////////////////////////////////////////////

		Model metamodel= null;
		try {
			metamodel= (Model)Model.readText
			           (new File("EXPRESS/frame_classes.exp"));
		}
		catch (FileNotFoundException fnfe) {
			System.err.println ("FileNotFoundException: "
			+ "EXPRESS/frame_classes.exp");
			System.exit(1);
		}
		catch (IOException ioe) {
			System.err.println(ioe.getMessage());
			System.exit(1);
		}
		catch (Throwable t) { 
			System.err.println(t);
			t.printStackTrace();
		}

		//////////////////////////////////////////////////////////
		// Create frame classes for entities and defined types  //
		// in the metamodel (needed for frame templates as well //
		// as the Jex frames themselves)                        //
		//////////////////////////////////////////////////////////

		ListIterator schemata= metamodel.getSchemas();
		while (schemata.hasNext()) {
			Schema schema= (Schema)schemata.next();

			FrameClass.createFrameClasses(schema);
			DefinedType.createDefinedTypes(schema);
		}

		///////////////////////////////////////////////////////////
		// Parse the frame templates in FIL-J/jex_to_frames.filj //
		///////////////////////////////////////////////////////////

		try {
			FileInputStream fis;
			FIL_J parser= new FIL_J(fis= new FileInputStream
			                        ("FIL-J/jex_to_frames.filj"));
			parser.Start();
			fis.close();			
		} catch (Exception e) {
			System.err.println("Frame.initialise: An exception of"
			+ " type " + e.getClass() + " occurred while parsing "
			+ "\"FIL-J/jex_to_frames.filj\"");
			System.err.print(e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}

		//////////////////////////////////////////////////////////
		// Convert the Jex objects from the metamodel to frames //
		//////////////////////////////////////////////////////////

		JexToFrames.convert(metamodel, false/*ALREADY DONE!*/);

		////////////////////////////////////////////////////////////
		// Slots with a tail remark of --IGNORE are to be ignored //
		// by the CCUS frame matching algorithm.                  //
		////////////////////////////////////////////////////////////

		Slot.ignoration();
	}

	/********************************************************************/
	// Frame constructor : void ->
	/*********************************************************************
	** Constructs this frame. This constructor should only be used when a 
	** frame's frame class has not been created yet - namely, when setting
	** up the EXPRESS metamodel). In all other circumstances, the
	** Frame(FrameClass) or Frame(String) constructors should be used.
	*********************************************************************/

	public Frame () {
		_fromInputModel= _markFramesAsInputModel;
	}

	/********************************************************************/
	// Frame constructor : FrameClass ->
	/*********************************************************************
	** Constructs this frame.
	** @param frameClass The frame class of the frame to be constructed,
	**  which specifies the set of slots the frame should have.
	*********************************************************************/
 
 	public Frame (FrameClass frameClass) {
		_frameClass= frameClass;
		_fromInputModel= _markFramesAsInputModel;
		try {
			frameClass.addInstance(this);
		} catch (InvalidFrameClass ifc) {
			System.err.println("Something has gone badly wrong "
			+ "for an InvalidFrameClass to be thrown here!");
		}
		_number= frameClass.getInstances().size();

		_numberOfFrames++;
	}
	
	/********************************************************************/
	// Frame constructor : String ->
	/*********************************************************************
	** Constructs this frame. The frame class is specified by name for
	** convenience.
	** @param frameClassName The name of the frame class of the frame to
	**                       be constructed, which specifies the set of
	**                       slots the frame should have.
	*********************************************************************/
 
 	public Frame (String frameClassName) {
		_frameClass= FrameClass.lookup(frameClassName);
		_fromInputModel= _markFramesAsInputModel;
		try {
			_frameClass.addInstance(this);
		} catch (InvalidFrameClass ifc) {
			System.err.println("Something has gone badly wrong "
			+ "for an InvalidFrameClass to be thrown here!");
		}
		_number= _frameClass.getInstances().size();

		_numberOfFrames++;
	}
	
	/********************************************************************/
	// Frame.getFrameClass : void -> FrameClass
	/*********************************************************************
	** Returns the frame class of this frame.
	** @return The frame class of this frame.
	*********************************************************************/

	public FrameClass getFrameClass () {
		return _frameClass;
	}

	/********************************************************************/
	// Frame.getNumber : void -> int
	/*********************************************************************
	** Returns this frame's instance number (unique to the frame class).
	** @return This frame's instance number.
	*********************************************************************/

	public int getNumber () {
		return _number;
	}
		
	/********************************************************************/
	// Frame.setNumber : int -> void
	/*********************************************************************
	** Gives this frame an instance number which is unique to this frame
	** class.
	** @param An instance number for this frame.
	*********************************************************************/

	public void setNumber (int number) {
		_number= number;
	}
		
	/********************************************************************/
	// Frame.getSlotFiller : Slot -> Object
	/*********************************************************************
	** Returns the filler of a given slot.
	** @param slot The slot whose filler you want to read.
	** @return The filler of `slot', as held in this frame.
	*********************************************************************/
 
 	public Object getSlotFiller (Slot slot) {
		return _frameRep.get(slot);
	}

	/********************************************************************/
	// Frame.getSlotFiller : String -> Object
	/*********************************************************************
	** Returns the filler of a given slot.
	** @param slot The name of the slot whose filler you want to read.
	** @return The filler of the slot, as held in this frame.
	** @exception InvalidSlot If the frame does not have a slot of that
	**  name (as specified in its frame class).
	*********************************************************************/

	public Object getSlotFiller (String slotName)
	                            throws InvalidSlot
	{
		return getSlotFiller(getFrameClass().getSlot(slotName));
	}

	/********************************************************************/
	// Frame.gsf : String -> Object
	/*********************************************************************
	** A quick and dirty version of getSlotFiller(String) for brevity.
	** Catches InvalidSlot exception within rather than throwing it out.
	** @param slot The name of the slot whose filler you want to read.
	** @return The filler of the slot, as held in this frame.
	*********************************************************************/

	public Object gsf (String slotName) {
		try {
			return getSlotFiller
			       (getFrameClass().getSlot(slotName));
		} catch (InvalidSlot is) {
			System.err.println(this + ".gsf: InvalidSlot \""
			+ slotName +"\"");
			System.exit(1); return null;
		}
	}

	/********************************************************************/
	// Frame.setSlotFiller : (Slot x Object) -> void
	/*********************************************************************
	** Fills a given slot with a given value. Records this slot as having
	** been written directly.
	** @param slot The slot you want to write into.
	** @param filler The filler (value) you want to write into `slot',
	**               which must be of the right type for `slot'.
	*********************************************************************/
 
	public void setSlotFiller (Slot slot, Object filler) {
		_slotsSetDirectly.put(slot,new Boolean(true));

		if (filler != null) {
			_frameRep.put(slot,filler);
		} else {
			_frameRep.remove(slot);
		}

		if (filler instanceof Frame) {
			setInverseSlot(slot,(Frame)filler);
			if (_caseProblemPart)
				((Frame)filler).markCaseProblemPart();
		}
		else if (filler instanceof Vector) {
			Enumeration fillers= ((Vector)filler).elements();
			while (fillers.hasMoreElements()) {
				Object nextFiller= fillers.nextElement();
				if (nextFiller instanceof Frame) {
					setInverseSlot(slot,(Frame)nextFiller);
					if (_caseProblemPart)
						((Frame)nextFiller).
							markCaseProblemPart();
				}
				else break;
			}
		}
	}

	/********************************************************************/
	// Frame.setInverseSlot : (Slot x Frame) -> void
	/*********************************************************************
	** If the given slot has an inverse slot, then set that inverse slot
	** to this frame for the given filler frame.
	** @param slot The slot being set.
	** @param filler The filler of the slot being set (a frame).
	** @param filler The filler (value) you want to write into `slot',
	**               which must be of the right type for `slot'.
	*********************************************************************/
 
	private void setInverseSlot (Slot slot, Frame filler) {
		Slot inverseSlot= Slot.lookupInverseSlot
		                      (slot.getFrameClassDeclaredIn().
		                            getName(),
		                       slot.getName(),
		                       filler.getFrameClass());

		if (inverseSlot != null) {
			if (filler.isSlotSetDirectly(inverseSlot))
				return;
			if (! inverseSlot.isAggType()) {
				filler.setSlotFillerIndirectly
				       (inverseSlot,this);
			} else {
				Vector agg= (Vector)filler.getSlotFiller
				                           (inverseSlot);
				if (agg == null) {
					if (inverseSlot.isOrderedAggType())
						agg= new OrderedVector();
					else
					if (inverseSlot.isUnorderedAggType())
						agg= new UnorderedVector();
					else {
						System.err.println("The slot "
						+ inverseSlot + " is an aggre"
						+ "gation but not a SET, BAG,"
						+ " LIST or ARRAY.");
						System.exit(1);
					}
				}
				agg.addElement(this);
				filler.setSlotFillerIndirectly
				       (inverseSlot,agg);
			}
		}
	}

	/********************************************************************/
	// Frame.setSlotFillerIndirectly : (Slot x Object) -> void
	/*********************************************************************
	** Fills a given slot with a given value indirectly (i.e. writing an
	** inverse slot in response to writing its `FOR' slot). Doesn't record
	** a direct write on this slot, and doesn't look for any inverses to
	** fill (because an inverse attribute cannot be the `FOR' attribute of
	** another inverse attribute).
	** @param slot The slot you want to write into.
	** @param filler The filler (value) you want to write into `slot',
	**               which must be of the right type for `slot'.
	*********************************************************************/

	private void setSlotFillerIndirectly (Slot slot, Object filler) {
		if (filler != null) {
			_frameRep.put(slot,filler);
		} else {
			_frameRep.remove(slot);
		}

		if (filler instanceof Frame) {
			if (_caseProblemPart)
				((Frame)filler).markCaseProblemPart();
		}
		else if (filler instanceof Vector) {
			Enumeration fillers= ((Vector)filler).elements();
			while (fillers.hasMoreElements()) {
				Object nextFiller= fillers.nextElement();
				if (nextFiller instanceof Frame) {
					if (_caseProblemPart)
						((Frame)nextFiller).
							markCaseProblemPart();
				}
				else break;
			}
		}
	}

	/********************************************************************/
	// Frame.setSlotFiller : (String x Object) -> void
	/*********************************************************************
	** Fills a given slot with a given value.
	** @param slotName The name of the slot you want to write into.
	** @param filler The filler (value) you want to write into the slot,
	**               which must be of the right type for the slot.
	** @exception InvalidSlot If the frame does not have that slot
	**                        (as specified in its frame class).
	*********************************************************************/
 
	public void setSlotFiller (String slotName, Object filler)
	                          throws InvalidSlot
	{
		setSlotFiller(getFrameClass().getSlot(slotName), filler);
	}

	/********************************************************************/
	// Frame.ssf : (String x Object) -> void
	/*********************************************************************
	** A quick and dirty version of setSlotFiller(String,Object) for
	** brevity. Catches InvalidSlot exception within rather than throwing
	** it out.
	** @param slotName The name of the slot you want to write into.
	** @param filler The filler (value) you want to write into the slot,
	**               which must be of the right type for the slot.
	*********************************************************************/
 
	public void ssf (String slotName, Object filler) {
		try {
			setSlotFiller(getFrameClass().getSlot(slotName),
			              filler);
		} catch (InvalidSlot is) {
			System.err.println(this + ".ssf: InvalidSlot \""
			+ slotName +"\"");
			System.exit(1);
		}
	}

	/********************************************************************/
	// Frame.addElementToSlotFiller : (Slot x Object) -> void
	/*********************************************************************
	** Adds an element to an aggregation slot filler.
	** @param aggSlot The aggregation slot you want to add an element to.
	** @param element The element you want to add to the slot filler.
	*********************************************************************/
 
	public void addElementToSlotFiller (Slot aggSlot, Object element) {
		Vector filler= null;
		try {
			filler= (Vector)getSlotFiller(aggSlot);
		} catch (ClassCastException cce) {
			System.err.println("Frame.addElementToSlotFiller: "
			+ aggSlot + " is not an aggregation slot!");
			System.err.println(cce);
			System.exit(1);
		}
		if (filler == null) {
			if (aggSlot.isOrderedAggType()) {
				filler= new OrderedVector();
			} else {
				filler= new UnorderedVector();
		}	}
		filler.addElement(element);
		setSlotFiller(aggSlot, filler);
	}

	/********************************************************************/
	// Frame.addElementToSlotFiller : (String x Object) -> void
	/*********************************************************************
	** Adds an element to an aggregation slot filler.
	** @param aggSlotName The name of the aggregation slot you want to add
	**                    an element to.
	** @param element The element you want to add to the slot filler.
	** @exception InvalidSlot If the frame does not have that slot
	**                        (as specified in its frame class).
	*********************************************************************/
 
	public void addElementToSlotFiller (String aggSlotName,Object element)
	                                   throws InvalidSlot
	{
		addElementToSlotFiller(getFrameClass().getSlot(aggSlotName),
		                       element);
	}

	/********************************************************************/
	// Frame.aesf : (String x Object) -> void
	/*********************************************************************
	** A quick and dirty version of addElementToSlotFiller(String,Object)
	** for brevity. Catches InvalidSlot exception within rather than
	** throwing it out.
	** @param aggSlotName The name of the aggregation slot you want to add
	**                    an element to.
	** @param element The element you want to add to the slot filler.
	*********************************************************************/
 
	public void aesf (String aggSlotName, Object element) {
		try {
			addElementToSlotFiller
			  (getFrameClass().getSlot(aggSlotName), element);
		} catch (InvalidSlot is) {
			System.err.println(this + ".aesf: InvalidSlot \""
			+ aggSlotName +"\"");
			System.exit(1);
		}
	}

	/********************************************************************/
	// Frame.deleteElementFromSlotFiller : (Slot x Object) -> void
	/*********************************************************************
	** Deletes an element from an aggregation slot filler.
	** @param aggSlot The aggregation slot you want to delete an element
	**                from.
	** @param element The element you want to delete from the slot filler.
	*********************************************************************/
 
	public void deleteElementFromSlotFiller (Slot aggSlot,Object element){
		Vector filler= null;
		try {
			filler= (Vector)getSlotFiller(aggSlot);
		} catch (ClassCastException cce) {
			System.err.println("Frame.deleteElementFromSlotFiller"
			+ ": " + aggSlot + " is not an aggregation slot!");
			System.err.println(cce);
			System.exit(1);
		}
		if (filler == null) {
			if (aggSlot.isOrderedAggType()) {
				filler= new OrderedVector();
			} else {
				filler= new UnorderedVector();
		}	}
		filler.removeElement(element);
		if (filler.size() == 0 && aggSlot.convertEmptyToNull())
			filler= null;
		setSlotFiller(aggSlot, filler);
	}

	/********************************************************************/
	// Frame.deleteElementFromSlotFiller : (String x Object) -> void
	/*********************************************************************
	** Deletes an element from an aggregation slot filler.
	** @param aggSlot The name of the aggregation slot you want to delete
	**                an element from.
	** @param element The element you want to delete from the slot filler.
	** @exception InvalidSlot If the frame does not have that slot
	**                        (as specified in its frame class).
	*********************************************************************/
 
	public void deleteElementFromSlotFiller (String aggSlotName,
	                                         Object element)
	                                        throws InvalidSlot
	{
		deleteElementFromSlotFiller
		  (getFrameClass().getSlot(aggSlotName), element);
	}

	/********************************************************************/
	// Frame.desf : (String x Object) -> void
	/*********************************************************************
	** A quick and dirty version of deleteElementFromSlotFiller(String,
	** Object) for brevity. Catches InvalidSlot exception within rather
	** than throwing it out.
	** @param aggSlot The name of the aggregation slot you want to delete
	**                an element from.
	** @param element The element you want to delete from the slot filler.
	*********************************************************************/
 
	public void desf (String aggSlotName, Object element) {
		try {
			deleteElementFromSlotFiller
			  (getFrameClass().getSlot(aggSlotName), element);
		} catch (InvalidSlot is) {
			System.err.println(this + ".desf: InvalidSlot \""
			+ aggSlotName +"\"");
			System.exit(1);
		}
	}

	/********************************************************************/
	// Frame.isSlotSetDirectly : Slot -> boolean
	/*********************************************************************
	** Returns true just in case a given slot in this frame was written
	** directly (i.e. not just filled in automatically as an inverse
	** slot).
	** @param slot A slot in this frame (presumably inverse).
	** @return Whether `slot' has been written directly in this frame.
	*********************************************************************/

	public boolean isSlotSetDirectly (Slot slot) {
		return _slotsSetDirectly.get(slot) != null;
	}

	/********************************************************************/
	// Frame.getSlots : void -> ListIterator
	/*********************************************************************
	** Returns an iterator over the filled slots in this frame, in the
	** order that they were filled. (If a slot was filled twice, its
	** position in the iterator is determined by the first time it was
	** filled.)
	** @return An iterator over the filled slots in this frame.
	*********************************************************************/

	public ListIterator getSlots () {
		return _frameRep.keyListIterator();
	}

	/********************************************************************/
	// Frame.markCaseProblemPart : void -> void
	/*********************************************************************
	** Marks this frame as being in the problem part of a case, recursing
	** for all frames in the slot fillers of this frame.
	*********************************************************************/

	public void markCaseProblemPart () {
		if (! _caseProblemPart) {
			_caseProblemPart= true;
		
			ListIterator filledSlots= getSlots();
			while (filledSlots.hasNext()) {
				Slot slot= (Slot)filledSlots.next();
				processSlotFiller(getSlotFiller(slot));
			}
		}
	}
	private void processSlotFiller (Object filler) {
		if (filler instanceof Frame)
			((Frame)filler).markCaseProblemPart();
		else if (filler instanceof Vector) {
			Enumeration elements= ((Vector)filler).elements();
			while (elements.hasMoreElements()) {
				processSlotFiller(elements.nextElement());
			}
		}
	}

	/********************************************************************/
	// Frame.isCaseProblemPart : void -> boolean
	/*********************************************************************
	** Returns true just in case this frame is in the `problem' part of a
	** case.
	** @return Whether this frame is in the `problem' part of a case.
	*********************************************************************/

	public boolean isCaseProblemPart () {
		return _caseProblemPart;
	}

	/********************************************************************/
	// Frame.isInputModelFrame : void -> boolean
	/*********************************************************************
	** Returns true just in case this frame is marked as being from an
	** input model (as opposed to from the frame class model, or part of
	** a case).
	** @return Whether this frame is from an input model.
	*********************************************************************/

	public boolean isInputModelFrame () {
		return _fromInputModel;
	}

	/********************************************************************/
	// Frame.toString : void -> String
	/*********************************************************************
	** Returns a string representing this frame. The format is
	** <tt><i>frame-class</i>#<i>number</i></tt>.
	** @return The string representation of this frame.
	*********************************************************************/

	public String toString () {
		return getFrameClass() + "#" + _number;
	}

	/********************************************************************/
	// Frame.print : PrintStream -> void
	/*********************************************************************
	** Prints the frame in human-readable text form. Any references are
	** just printed as they are.
	** @param ps A stream to print the frame to.
	*********************************************************************/

	public void print (PrintStream ps) {
		Slot slot;

		generateFIL(new PrintWriter(ps,true), false, false);
		ps.println();
	}
	
	/********************************************************************/
	// static Frame.generateFIL : boolean -> void
	/*********************************************************************
 	** Writes all frames in the system to a set of FIL files.
	** @param HTML If true, files will be augmented with HTML tags and
	**             have the extension <TT>.html</TT> instead of
	**             <TT>.fil</TT>.
	** @param suppress If true, only input model frames will be included
	**                 in the file (i.e. frame class model frames and
	**                 frames that are part of cases will be excluded).
	** @param indexHTML The PrintWriter for index.html (optional).
	*********************************************************************/

	public static void generateFIL (boolean HTML, boolean suppress,
	                                PrintWriter indexHTML)
	{
		Enumeration frameClasses= FrameClass.getAllFrameClasses();
		while (frameClasses.hasMoreElements()) {
			FrameClass fc= (FrameClass)frameClasses.nextElement();

			if (fc.anyInputModelFrames()
			 || (! suppress) && (fc.getInstances().size() > 0))
			{
				fc.generateFILforInstances(HTML,suppress,
				                           indexHTML);
			}
		}
	}

	/********************************************************************/
	// Frame.generateFIL : PrintWriter x boolean -> void
	/*********************************************************************
	** Writes the frame into an FIL file. Any references are given
	** corresponding hypertext links if the HTML flag is true.
	** @param pw A file to generate the HTML into (with automatic line
	**  flushing enabled).
	** @param HTML If true, file will be augmented with HTML tags.
	** @param suppress If true, only input model frames will be included
	**                 in the file (i.e. links to the frames representing
	**                 the frame classes and slots will be suppressed).
	*********************************************************************/

	public void generateFIL (PrintWriter pw, boolean HTML,
	                         boolean suppress)
	{
		Slot slot;
		Object filler;

		////////////
		// Header //
		////////////
		if (HTML) pw.println("<A NAME=\"" + _number +"\"></A>");
		if (HTML) pw.println("<PRE>");
		if (HTML && _caseProblemPart) pw.println("<I>");
		pw.print("INSTANCE ");
		if (HTML && !suppress) pw.print("<A HREF=\"entity_type.html#"
		+ getFrameClass().getClassFrame().getNumber() + "\">");
		pw.print(getFrameClass());
		if (HTML && !suppress) pw.print("</A>");
		pw.println("#" + _number + ";");

		///////////
		// Slots //
		///////////
		Enumeration slots= getFrameClass().getSlots().elements();
		while (slots.hasMoreElements()) {
			slot= (Slot)slots.nextElement();
			filler= getSlotFiller(slot);
			if (filler != null) {
			 	pw.print("\t");
				if (HTML && !suppress) pw.print("<A HREF=\""
				+ slot.getSlotFrame().getFrameClass()
				+ ".html#" + slot.getSlotFrame().getNumber()
				+ "\">");
				pw.print(slot.getName());
				if (HTML && !suppress) pw.print("</A>");
				pw.print(" = ");
				if (filler instanceof FILisable)
			 		pw.print(((FILisable)filler).
					                     FILise(HTML));
			 	else
			 		pw.print(filler);
				pw.print(";");
				if (! isSlotSetDirectly(slot)) {
					pw.print(" ");
					if (HTML) pw.print("<EM>");
					pw.print("(* set indirectly *)");
					if (HTML) pw.print("</EM>");
				}
				pw.println();
			}
		}		
		pw.println("END_INSTANCE;");
		if (HTML && _caseProblemPart) pw.println("</I>");
		if (HTML) pw.println("</PRE>");
	}

	/********************************************************************/
	// Frame.FILise : boolean -> String
	/*********************************************************************
	** Returns an FIL reference to this frame.
	** @param HTML If true, the FIL will be augmented with HTML tags.
	** @return A string representing the frame reference in FIL.
	*********************************************************************/

	public String FILise (boolean HTML) {
		String returnValue= "";
		if (HTML)
			returnValue= "<A HREF=\"" + getFrameClass()
		                   + ".html" + "#" + getNumber() + "\">";
		returnValue+= toString();

		if (HTML)
			returnValue+= "</A>";
		return returnValue;
	}
	
	/********************************************************************/
	// static Frame.remove : Frame -> void
	/*********************************************************************
 	** Removes a frame from the index of its frame class.
	** @param frame The frame to remove.
	*********************************************************************/

	public static void remove (Frame frame) {
		try {
			frame.getFrameClass().removeInstance(frame);
		} catch (InvalidFrameClass ifc) {
			System.err.println("Something has gone badly wrong "
			+ "for an InvalidFrameClass to be thrown here!");
		}
	}
}
