      /////////////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM          //
    //  (Version 2)                                 ///
   // Written by Andrew Broad                      ////
  //                                              /////
 // InvocationList.java (1999:11:25 12:23)       //////
//////////////////////////////////////////////////////
// Methods of InvocationList:                    ////
// - InvocationList constructor()                ///
// - invoke                                      //
//////////////////////////////////////////////////

import java.io.*;
import java.util.*;

import jex.*;

/*****************************************************************************
** The class `InvocationList' represents a list of method calls (and may later
** include fields). It is a subclass of Vector, and adds a method to perform
** the invocations in order.
*****************************************************************************/

public class InvocationList extends Vector {
	/********************************************************************/
	// InvocationList constructor : void ->
	/*********************************************************************
	** Constructs an empty invocation list.
	*********************************************************************/
 
 	public InvocationList () {
		super();
	}
	
	/********************************************************************/
	// InvocationList.invoke : Object -> Object
	/*********************************************************************
	** Performs the invocations in the list in order on a given object,
	** returning the final result.
	** @param object The object on which the invocations are to be
	**               performed.
	** @return The final result of the invocations.
	*********************************************************************/

	public Object invoke (Object object) {
//System.err.print("Invoking InvocationList: [");
//Enumeration e= elements();
//while (e.hasMoreElements()) {
//System.err.print(e.nextElement());
//if (e.hasMoreElements()) System.err.print(",");
//}
//System.err.println("]");
		Enumeration invocations= elements();
		while (invocations.hasMoreElements()) {
			MethodCall invocation= (MethodCall)
			                       invocations.nextElement();
			if (object != null) {
				object= invocation.invoke(object);
			} else {
				System.err.print("Warning: null." +
				  invocation.getMethodName() + "(");
				Object[] a= invocation.getActualParameters();
				for (int i= 0; i < a.length; i++) {
					System.err.print(a[i]);
					if (i < a.length - 1)
						System.err.print(",");
				}
				System.err.println(") not invoked");
			}
		}
//System.err.println();
		return object;
	}
}
