      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // MethodCall.java (1999:07:27 10:36)  //////
//////////////////////////////////////////////
// Methods of MethodCall:               //////
// - MethodCall constructor             //////
// - invoke                             /////
// - getMethodName                      ////
// - getParameterTypes                  ///
// - getActualParameters                //
/////////////////////////////////////////

import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import jex.*;

/*****************************************************************************
** The class `MethodCall' represents a method call in a frame template.
*****************************************************************************/

public class MethodCall implements Serializable {
	//private String _objectIdentifier;
	private String _methodName;
	private Class[] _parameterTypes;
	private Object[] _actualParameters;

	/********************************************************************/
	// MethodCall constructor : String x Vector ->
	/*********************************************************************
	** Constructs this method call.
	** @param methodName The name of the method to be called.
	** @param actualParameters The actual parameters of the method call.
	*********************************************************************/
 
 	public MethodCall (String methodName, Vector actualParameters) {
		_methodName= methodName;
		_actualParameters= new Object[actualParameters.size()];
		for (int n= 0; n < actualParameters.size(); n++) {
			Object p= actualParameters.elementAt(n);
			_actualParameters[n]= p;
			_parameterTypes[n]= p.getClass();
		}
	}
	
	/********************************************************************/
	// MethodCall.invoke : Object -> Object
	/*********************************************************************
	** Returns the name of the method to be called.
	** @param object The object on which the method is to be invoked.
	** @return The value returned by the method call.
	*********************************************************************/

	public Object invoke (Object object) {
//System.err.print("Invoking " + this + ": " + object + "." + _methodName + "(");
//for (int n= 0; n < _actualParameters.length; n++) {
//System.err.print(_parameterTypes[n] + " " + _actualParameters[n]);
//if (n < _actualParameters.length-1) System.err.println(",");
//}
//System.err.println(")");
		Class klass= object.getClass();
		Method method= null;
		try {
			method= klass.getMethod(getMethodName(),
			                        getParameterTypes());
		} catch (NoSuchMethodException nsme) {
			System.err.println("MethodCall.invoke: No such method"
			+ " as " + getMethodName());
			System.exit(1);
		}
		try {
			return method.invoke(object,getActualParameters());
		} catch (InvocationTargetException e) {
			System.err.print(object + "." + _methodName + "(");
			for (int i= 0; i < _actualParameters.length; i++) {
				System.err.print(_actualParameters[i]);
				if (i < _actualParameters.length - 1)
					System.err.println(",");
			}
			System.err.print(") threw a ");
			System.err.println(e.getTargetException().getClass());
			//e.printStackTrace();
			return null;
		} catch (Exception e) {
			System.err.println("MethodCall.invoke: Exception of "
			+ "type " + e.getClass() + " caught.");
			System.err.println(e.getMessage());
			System.exit(1); return null;
		}
	}

	/********************************************************************/
	// MethodCall.getMethodName : void -> String
	/*********************************************************************
	** Returns the name of the method to be called.
	** @return The method name.
	*********************************************************************/

	public String getMethodName () {
		return _methodName;
	}

	/********************************************************************/
	// MethodCall.getParameterTypes : void -> Class[]
	/*********************************************************************
	** Returns the list of parameter types for this method call - as an
	** array of classes.
	** @return The array of parameter types for this method call.
	*********************************************************************/

	public Class[] getParameterTypes () {
		return _parameterTypes;
	}

	/********************************************************************/
	// MethodCall.getActualParameters : void -> Object[]
	/*********************************************************************
	** Returns the list of actual parameters for this method call - as an
	** array of objects.
	** @return The array of actual parameters of this method call.
	*********************************************************************/

	public Object[] getActualParameters () {
		return _actualParameters;
	}
}
