      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // Summary.java (2002:07:22 13:54)     /////
////////////////////////////////////////////
// Methods of Summary:                 ////
// - addEntityPair                     ///
// - generateFile                      //
////////////////////////////////////////

import java.io.*;
import java.util.*;

/*****************************************************************************
** The class `Summary' synthesises the file "summary.html" for a CCUS run.
*****************************************************************************/

public class Summary {
	private static Vector/*EntityPair*/ _entityPairs= new Vector();

	/********************************************************************/
	// static Summary.addEntityPair : EntityPair -> void
	/*********************************************************************
	** Add an entity-pair to the summary.
	** @param entityPair The entity-pair to add.
	*********************************************************************/

	public static void addEntityPair (EntityPair entityPair) {
		_entityPairs.add(entityPair);
	}

	/********************************************************************/
	// static Summary.generateFile : PrintWriter x String -> void
	/*********************************************************************
	** Generates summary.html.
	** @param summaryFile The PrintWriter object for summary.html.
	** @param name The name of the experiment.
	** @param frameNetwork1 The top-level frame of the first model.
	** @param frameNetwork2 The top-level frame of the second model.
	*********************************************************************/

	public static void generateFile (PrintWriter summaryFile, String name,
	                             Frame frameNetwork1, Frame frameNetwork2)
	{
		////////////
		// Header //
		////////////
		summaryFile.println("<HTML><TITLE>" + name + "</TITLE>");
		summaryFile.println("<H1><A HREF=\"../../../\">Andrew Broad's"
		+ " Website</A><BR>");
		summaryFile.println("<A HREF=\"../../\">Computer Science</A>"
		+ "<BR>");
		summaryFile.println("<A HREF=\"../\">PhD Project</A><BR>");
		summaryFile.println("<A HREF=\"../results.html\">Experimental"
		+ " Results</A><BR>");
		summaryFile.println("<A HREF=\"./\">" + name + "</A><BR>");
		summaryFile.println("Summarised Output</H1>");
		summaryFile.println();

		////////////////////////////
		// Entity-correspondences //
		////////////////////////////
		summaryFile.println("<UL>");
		ListIterator entityPairs= _entityPairs.listIterator();
		while (entityPairs.hasNext()) {
			((EntityPair)entityPairs.next()).
			             printHeaderLine(summaryFile);
		}
		summaryFile.println("</UL>");

		////////////////////////////////
		// Entity-non-correspondences //
		////////////////////////////////
		summaryFile.println("<UL>");
		Frame frame;
		ListIterator schemata1= ((Vector)frameNetwork1.gsf("schemata")
		                        ).listIterator();
		while (schemata1.hasNext()) {
			ListIterator entities1= ((Vector)((Frame)schemata1.
			                        next()).gsf("entities")).
			                        listIterator();
			while (entities1.hasNext()) {
				if ((frame= (Frame)entities1.next()).
				    gsf("corresponds_to") == null)
				{
					summaryFile.println("<LI><A HREF=\""
					+ frame.getFrameClass() + ".html#"
					+ frame.getNumber() + "\">"
					+ "<TT>" + frame.gsf("name") + "</TT>"
					+ "</A> doesn't correspond to anythin"
					+ "g in the second model</LI>");
				}
			}
		}
		summaryFile.println("</UL>");

		summaryFile.println("<UL>");
		ListIterator schemata2= ((Vector)frameNetwork2.gsf("schemata")
		                        ).listIterator();
		while (schemata2.hasNext()) {
			ListIterator entities2= ((Vector)((Frame)schemata2.
			                        next()).gsf("entities")).
			                        listIterator();
			while (entities2.hasNext()) {
				if ((frame= (Frame)entities2.next()).
				    gsf("corresponds_to") == null)
				{
					summaryFile.println("<LI><A HREF=\""
					+ frame.getFrameClass() + ".html#"
					+ frame.getNumber() + "\">"
					+ "<TT>" + frame.gsf("name") + "</TT>"
					+ "</A> doesn't correspond to anythin"
					+ "g in the first model</LI>");
				}
			}
		}
		summaryFile.println("</UL>");

		//////////////////
		// Entity-pairs //
		//////////////////
		entityPairs= _entityPairs.listIterator();
		while (entityPairs.hasNext()) {
			summaryFile.println("<HR>");
			((EntityPair)entityPairs.next()).
			             generateOutput(summaryFile);
		}

		////////////
		// Footer //
		////////////
		summaryFile.println("<HR>");
		summaryFile.println("<A HREF=\"../../../email.html\">"
		+ "Email me</A>");
		summaryFile.println("</HTML>");
	}
}
