      ////////////////////////////////////////
     // THE CONSTRAINT-UNDERSTANDING SYSTEM //
    //  (Version 2)                        ///
   // Written by Andrew Broad             ////
  //                                     /////
 // Tabulate.java (2002:12:02 14:55)    //////
//////////////////////////////////////////////
// Methods of Tabulate:                ///////
// - main                              //////
// - tabulate                          /////
// - processArguments                  ////
// - processFlags                      ///
// - usage                             //
////////////////////////////////////////

import java.io.*;
import java.util.*;

/*****************************************************************************
** The class `Tabulate' extracts experiment data from the CCUS output files
** out[1-8].txt, and synthesises a LaTeX table.
*****************************************************************************/

public class Tabulate implements Serializable {
	public static boolean _changebars= false;
	public static boolean _smallTable= false;
	public static boolean _frameClassCorrespondences= false;
	public static Hashtable _fcCorrsTable/*String->Array<Integer>*/
	= new Hashtable();

	/********************************************************************/
	// static Tabulate.main : void -> void
	/*********************************************************************
	** Synthesise a LaTeX table from CCUS outputs.
	*********************************************************************/

	public static void main (String argv[]) {
		int tableNumber= processArguments(argv);

		int[][] table= new int[8][11];

		int modes= _smallTable ? 3 : 8;
		if (_smallTable) {
			table[0]= tabulate("out6.txt", 0);
			table[1]= tabulate("out7.txt", 1);
			table[2]= tabulate("out1.txt", 2);	
		} else {
			for (int e= 0; e < 8; e++) {
				table[e]= tabulate("out" + (e+1) + ".txt", e);
			}
		}
		System.out.println("\\begin{table}");
		if (_changebars)
			System.out.print("\\cbstart");
		System.out.println("\\begin{center}");
		System.out.println("\\begin{tabular}");
		if (_smallTable)
	 		System.out.println("{|c||c|c|c|}");
		else
			System.out.println("{|c||c|c|c|c|c|c|c|c|}");
		System.out.println("\\hline");
		if (_smallTable)
			System.out.println("Remembrance heuristic            "
		+ "& NR & M & IR \\\\");
		else
			System.out.println("Mode                             "
			+ "& E & -T & -U & -S & -H & -R & M & X \\\\");
		System.out.println("\\hline\\hline");
		System.out.print("HLCs extracted from first model  ");
		for (int i= 0; i < modes; i++) {
			System.out.print("& " + table[i][2] + " ");
		}
		System.out.println("\\\\");
		System.out.println("\\hline");
		System.out.print("HLCs extracted from second model ");
		for (int i= 0; i < modes; i++) {
			System.out.print("& " + table[i][3] + " ");
		}
		System.out.println("\\\\");
		System.out.println("\\hline");
		System.out.print("Comparisons                      ");
		for (int i= 0; i < modes; i++) {
			System.out.print("& " + table[i][4] + " ");
		}
		System.out.println("\\\\");
		System.out.println("\\hline");
		System.out.print("HLC-comparisons                  ");
		for (int i= 0; i < modes; i++) {
			System.out.print("& " + table[i][5] + " ");
		}
		System.out.println("\\\\");
		System.out.println("\\hline");
		System.out.print("Correspondences                  ");
		for (int i= 0; i < modes; i++) {
			System.out.print("& " + table[i][6] + " ");
		}
		System.out.println("\\\\");
		System.out.println("\\hline");
		System.out.print("HLC-correspondences              ");
		for (int i= 0; i < modes; i++) {
			System.out.print("& " + table[i][7] + " ");
		}
		System.out.println("\\\\");
		System.out.println("\\hline");
		System.out.print("Differences                      ");
		for (int i= 0; i < modes; i++) {
			System.out.print("& " + table[i][8] + " ");
		}
		System.out.println("\\\\");
		System.out.println("\\hline");
		if (! _smallTable) {
			System.out.print("FC-thesaurus lookups             ");
			for (int i= 0; i < 8; i++) {
				System.out.print("& " + table[i][9] + " ");
			}
			System.out.println("\\\\");
			System.out.println("\\hline");
			System.out.print("Name-thesaurus lookups           ");
			for (int i= 0; i < 8; i++) {
				System.out.print("& " + table[i][10] + " ");
			}
			System.out.println("\\\\");
			System.out.println("\\hline");
		}
		System.out.println("\\end{tabular}");
		System.out.print("\\end{center}");
		if (_changebars)
			System.out.print("\\cbend{}");
		System.out.print("\n\\caption{Experimental results for the "
		+ "models in ");
		if (_smallTable)
			System.out.println("Figure~\\ref{fig:7-"
			+ (tableNumber+4) + "}.}");
		else
			System.out.println("Section~\\ref{sec:F."
			+ tableNumber + "}.}");

		System.out.println("\\label{table" + (_smallTable ? "7" : "F")
		                 + "-" + tableNumber + "}");
		System.out.println("\\end{table}");

		/////////////////////////////////
		// Frame-class correspondences //
		/////////////////////////////////
		if (_frameClassCorrespondences) {
			System.out.println("\n\n\\begin{table}");
			System.out.println("\\begin{center}");
			System.out.println("\\begin{tabular}");
			System.out.println("{|c||c|c|c|c|c|c|c|c|}");
			System.out.println("\\hline");
			System.out.println("Mode                             "
			+ "& E & -T & -U & -S & -H & -R & M & X \\\\");
			System.out.println("\\hline\\hline");

			Enumeration frameClasses= _fcCorrsTable.keys();
			while (frameClasses.hasMoreElements()) {
				String fcName= (String)frameClasses.
				                       nextElement();
				System.out.print("\\texttt{" + fcName
				+"} ");
				for (int i= 0; i < 8; i++) {
					System.out.print("& "
					+ ((Vector)_fcCorrsTable.get(fcName)).
					  elementAt(i) + " ");
				}
				System.out.println("\\\\");
				System.out.println("\\hline");
			}
			System.out.println("\\end{tabular}");
			System.out.print("\\end{center}");
			System.out.print("\n\\caption{Breakdown of correspond"
			+ "ences for the models in ");
			System.out.println("Section~\\ref{sec:F."
			+ tableNumber + "}.}");

			System.out.println("\\label{tableF-"
			+ (tableNumber + 1) + "}");
			System.out.println("\\end{table}");
		}
	}

	/********************************************************************/
	// static Tabulate.tabulate : void -> void
	/*********************************************************************
	** Extract experiment data from a CCUS output file.
	** @param filename The filename of the CCUS output file.
	** @param col The column (mode) in the table.
	** @return An array of data extracted from said file.
	*********************************************************************/

	public static int[] tabulate (String filename, int col) {
		int[] expt= new int[11]; int e= 0;

		BufferedReader in= null; FileInputStream fis= null;
		try {
			in= new BufferedReader(new InputStreamReader
			         (fis= new FileInputStream(filename)));
		} catch (FileNotFoundException fnfe) {
			System.err.println("Tabulate.tabulate: FileNotFoundEx"
			+ "ception for \"" + filename + "\"");
			System.exit(1);
		}
		String line;
		int i,n;
		char c;
		try {
			while ((line= in.readLine()) != null) {
				try {
					if (line.substring(0,6).
					    equals("Number"))
					{
						i= 6; // start looking for ":"
						while (line.charAt(i) != ':'){
							i++;
						}
						i+= 2; // skip ": "
						n= Integer.parseInt
						   (line.substring(i,
						         line.length()));
						expt[e++]= n;
					}
					/////////////////////////////////
					// Frame-class correspondences //
					/////////////////////////////////
					if (line.substring(0,11).
					    equals("=Number of "))
					{
						String fcName= "";
						for (i= 11; //start of FC-name
						     line.charAt(i) != '-';
						     i++)
						{
							if (line.charAt(i)
							    == '_') {
								fcName+= "\\";
							}
							fcName
							+= line.charAt(i);
						}
						i+= 18;
						// skip "-correspondences: "
						n= Integer.parseInt
						   (line.substring(i,
						         line.length()));
						Vector row= (Vector)
						    _fcCorrsTable.get(fcName);
						if (row == null) {
							row= new Vector();
							for (int j= 0; j < 8;
							     j++) {
								row.add(j, new
								  Integer(0));
							}
							_fcCorrsTable.
							  put(fcName, row);
						}
						row.remove(col);
						row.add(col, new Integer(n));
					}
				} catch (IndexOutOfBoundsException ioobe) {}
			}
			fis.close();
		} catch (IOException ioe) {
			System.err.println("Tabulate.tabulate: IOException th"
			+ "rown by BufferedReader.readLine()");
			System.exit(1);
		}
		return expt;
	}

	/********************************************************************/
	// static Tabulate.processArguments : String[] -> int
	/*********************************************************************
	** Method to process the command-line arguments.
	** @param argv The vector of command-line arguments. There should be
	**             exactly one table-number, and any number of -flags:
	**             -c: insert change-bars
	**             -s: generate small table for Section 7.2.4
	**             -f: include correspondences by frame-class
	** @return The table-number.
	*********************************************************************/

	private static int processArguments (String argv[]) {
                String tableNumber= null;
		for (int i= 0; i < argv.length; i++) {
			if (argv[i].startsWith("-")) {
				processFlags(argv[i]);
			}
			else {
				if (tableNumber == null)
					tableNumber= argv[i];
				else usage();
			}
		}
		if (tableNumber == null) usage();
		return Integer.parseInt(tableNumber);
	}

	/********************************************************************/
	// static Tabulate.processFlags : String[] -> void
	/*********************************************************************
	** A subroutine to process the flags in a command-line argument
	** starting with `-'. This method assumes that it is only called with
	** strings starting with `-', and ignores the first character.
	** @param arg A `flags' command-line argument (i.e. `-' followed by
	**            letters).
	*********************************************************************/

	private static void processFlags (String arg) {
		char[] flags= arg.toCharArray();

		for (int i= 1; i < flags.length; i++) {
			switch (flags[i]) {
				case 'c':
					_changebars= true;
					break;
				case 's':
					_smallTable= true;
					break;
				case 'f':
					_frameClassCorrespondences= true;
					break;
				default:
					System.err.println("Warning: -"
					+ flags[i] + " not recognised.");
			}
		}
	}

	/********************************************************************/
	// static Tabulate.usage : void -> void
	/*********************************************************************
	** Print the usage report when the command-line arguments are invalid,
	** and exit the system.
	*********************************************************************/

	private static void usage () {
		System.err.println("Usage: java Tabulate <table_number>");
		System.err.println("    Flags:");
		System.err.println("-c: insert change-bars");
		System.err.println("-s: generate small table for "
		                 + "Section 7.2.4");
		System.err.println("-f: include correspondences by "
		                 + "frame-class");
		System.exit(1);
	}
}
